#include <stdio.h>
#include <stdlib.h>
#include <string.h>
 #include <sys/time.h>
#include <cuda_runtime_api.h>
#include <cuda.h>
#include <math.h>

#define REPRODUCE_TEST
static void minMaxDiff(float *bpcpu, float *bpgpu, float *projarr, int num, int kstart,
                       int support, int width);
double walltime(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return ((double)tv.tv_sec + tv.tv_usec / 1000000.);
}

__constant__ float cosines[256];
__constant__ float sines[256];

// declare texture reference for 2D float texture
texture<float, 2, cudaReadModeElementType> projtex;

__global__ void bpLine1D(float *slice, int pitch, int width, int j, 
                         float kproj, float xpstart, float costh) 
{
  unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
  if (i < width) {
    float xp = xpstart + i * costh + 0.5f;
    slice[j * pitch + i] += tex2D(projtex, xp, kproj + 0.5f);
  }
}

__global__ void bpLine2D(float *slice, int pitch, int width, int thick, 
                         float kproj, float xstart, float costh, float sinth) 
{
  unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
  unsigned int j = blockIdx.y * blockDim.y + threadIdx.y;
  int yy = j - thick / 2;
  
  if (i < width && j < thick) {
    float xp = xstart + i * costh + 0.5f - sinth * yy;
    slice[j * pitch + i] += tex2D(projtex, xp, kproj + 0.5f);
  }
}

__global__ void bpAllAngles(float *slice, int pitch, int width, int thick, int numproj,
                            int support)
{
  int k;
  float sum = 0.;
  unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
  unsigned int j = blockIdx.y * blockDim.y + threadIdx.y;
  int yy = j - thick / 2;
  if (i < width && j < thick) {
    for (k = 0; k < numproj; k++) {
      float xp = (i - width / 2.) * cosines[k] + support / 2. + 0.5f - sines[k] * yy;
      float kproj = k + 0.5f;
      sum += tex2D(projtex, xp, kproj);
    }
    slice[j * pitch + i] = sum;
  }
}

int main(int argc, char **argv)
{
  int current_device;
  int device_count = 0;
  float gflops;
  size_t width = 4096;
  size_t thick = 300;
  size_t sizetmp;
  int seed = 1234567;
  double maxang = 75.;
  double anginc = 1.;
  double DTOR = 3.14159 / 180.;
  int support, numproj, lastend, rad, xcp, xc, i, j, k, ix, sliceind, lineind, yy;
  int kstart, kend;
  float *projarr, *bpcpu, *bpgpu;
  float radlim, yc, proj, den, xstart, xx, fx;
  double theta, tstart;
  float costh, sinth;
  cudaError_t err;  
  struct cudaDeviceProp device_properties, best_properties;
  int max_gflops_device = -1;
  float max_gflops = 0;
  float cosarr[256], sinarr[256];

  if (argc > 1)
    width = atoi(argv[1]);
  if (argc > 2)
    thick = atoi(argv[2]);
  if (argc > 3)
  maxang = atof(argv[3]);
  if (argc > 4)
    anginc = atof(argv[4]);

  cudaGetDeviceCount( &device_count );
  for (current_device = 0; current_device < device_count; current_device++) {
    cudaGetDeviceProperties( &device_properties, current_device );
    gflops = device_properties.multiProcessorCount * device_properties.clockRate;
    if( gflops > max_gflops ) {
      max_gflops = gflops;
      max_gflops_device = current_device;
      best_properties = device_properties;
    }
  }
    
  if (max_gflops_device >= 0) {
#ifndef REPRODUCE_TEST
    printf("Best device %d, gflops %f, multiprocessors %d,  memory %d\n", max_gflops_device,
           gflops, best_properties.multiProcessorCount, best_properties.totalGlobalMem);
#endif
  } else {
    printf("No devices found\n");
    exit(0);
  }

  if (cudaSetDevice(max_gflops_device) != cudaSuccess) {
    printf("Error selecting device");
    exit(1);
  }

  // Get slices
  sizetmp = sizeof(float) * width * thick;
  bpcpu = (float *)malloc(sizetmp);
  bpgpu = (float *)malloc(sizetmp);
  for (i = 0; i < width * thick; i++) {
    bpcpu[i] = 0.;
    bpgpu[i] = 0.;
  }

  // Make the model projections
  support = ((((int)sqrt((double)(width * width + thick * thick))) / 2) + 1) * 2;
  numproj = 2 * (int)(maxang / anginc) + 1;
  sizetmp = sizeof(float) * support * numproj;
  projarr = (float *)malloc(sizetmp);
#ifndef REPRODUCE_TEST
  printf("width %d  thickness %d  max angle %.1f  increment %.1f  # angles %d\n", width,
         thick, maxang, anginc, numproj);
#endif
  srand(seed);
  radlim = thick / 4;
  lastend = 0;
  while (lastend < width - 2 * radlim - 2) {
    rad = (int)(rand() * radlim / RAND_MAX);
    if (rad < 3)
      continue;
    xc = lastend + rad + 1;
    yc = rad + ((thick - 2. * rad) * rand()) / RAND_MAX;
    den = 100. * rand() / RAND_MAX;
    for (j = 0; j < numproj; j++) {
      theta = DTOR * (j - numproj / 2) * anginc;
      xcp = (int)((xc - support / 2) * cos(theta) + (yc - thick / 2) * sin(theta))
        + support / 2;
      for (i = -rad; i <= rad; i++) {
        proj = (2. * sqrt((double)(rad * rad - i * i)) + 1.) * den;
        projarr[support * j + xcp + i] += proj;
        if (support * j + xcp + i < 0 || support * j + xcp + i >= support * numproj) 
          printf("Bad!!!\n");
      }
    }

    lastend += 2 * rad + 2;
  }

  // Do cpu backproj
  tstart = walltime();
  kstart = 0 ;
  kend = numproj - 1;
  for (k = kstart; k <= kend; k++) {
    theta = DTOR * (k - numproj / 2) * anginc;
    costh = cos(theta);
    sinth = sin(theta);
    cosarr[k] = costh;
    sinarr[k] = sinth;
    lineind = support * k;
    for (j = 0; j < thick; j++) {

      // NOTE when yy was float it assigned it wrong
      yy = j - thick / 2;
      xstart = - costh * width / 2. - sinth * yy + support / 2.;
      sliceind = width * j;
      for (i = 0; i < width; i++) {
        xx = xstart + i * costh;
        ix = (int)xx;
        fx = xx - ix;
        bpcpu[sliceind + i] += 
          (1. - fx) * projarr[lineind + ix] + fx * projarr[lineind + ix + 1];
      }
    }
  }

#ifndef REPRODUCE_TEST
  printf("CPU walltime %.4f sec\n", walltime() - tstart);
#endif
  // Allocate memory for slice on device
  float* devSlice;
  size_t slicePitch;
  sizetmp = width * sizeof(float);
  if (cudaMallocPitch((void **)&devSlice, &slicePitch, sizetmp, thick) != 
  //if (cudaMalloc((void **)&devSlice, sizetmp * thick) != 
      cudaSuccess) {
    printf("Failed to allocate slice array on device\n");
    exit(1);
  }

  // Allocate memory for projection lines and copy
  cudaChannelFormatDesc projDesc = cudaCreateChannelDesc
    (32, 0, 0, 0, cudaChannelFormatKindFloat);
  cudaArray* devProj;
  if (cudaMallocArray(&devProj, &projDesc, support, numproj) != cudaSuccess) {
    printf("Failed to allocate projection array on device\n");
    exit(1);
  }
  sizetmp = sizeof(float) * support * numproj;
  tstart = walltime();
  if (cudaMemcpyToArray(devProj, 0, 0, projarr, sizetmp, cudaMemcpyHostToDevice)
      != cudaSuccess) {
    printf("Failed to copy projection array to device\n");
    exit(1);
  }
#ifndef REPRODUCE_TEST
  printf("projection copy walltime %.4f sec\n", walltime() - tstart);
#endif

  // set texture parameters
  projtex.addressMode[0] = cudaAddressModeClamp;
  projtex.addressMode[1] = cudaAddressModeClamp;
  projtex.filterMode = cudaFilterModeLinear;
  projtex.normalized = false;
  
  // Bind the array to the texture
  if (cudaBindTextureToArray(projtex, devProj, projDesc) != cudaSuccess) {
    printf("Failed to bind projection array to texture\n");
    exit(1);
  }

  // Clear out the slice
  if (cudaMemset2D(devSlice, slicePitch, 0, width * sizeof(float), thick) !=
  //if (cudaMemset(devSlice, 0, width * sizeof(float) * thick) !=
     cudaSuccess) {
  //sizetmp = sizeof(float) * width;
  //if (cudaMemcpy2D(devSlice, sizetmp, bpgpu, slicePitch, sizetmp,  thick, 
                 //if (cudaMemcpy(devSlice, bpgpu, sizetmp * thick, 
    //                 cudaMemcpyHostToDevice) != cudaSuccess) {
    printf("Failed to clear slice array on device\n");
    exit(1);
  }
  tstart = walltime();

  // Loop on simple 1D kernel calls
  dim3 blockSize(best_properties.maxThreadsPerBlock / 2, 1, 1);
  dim3 gridSize((width + blockSize.x - 1) / blockSize.x, 1, 1);
  for (k = kstart; k <= kend; k++) {
    theta = DTOR * (k - numproj / 2) * anginc;
    costh = cos(theta);
    sinth = sin(theta);
    for (j = 0; j < thick; j++) {
      yy = j - thick / 2;
      xstart = - costh * width / 2. - sinth * yy + support / 2.;
      xx = k;
      // bpLine1D<<<1, width>>>(devSlice, slicePitch, j, (float)k, xstart, costh);
      bpLine1D<<<gridSize, blockSize>>>(devSlice, slicePitch / 4, width, j, xx, xstart,
                                        costh);
    }
  }
  err = cudaGetLastError();
  if (err != cudaSuccess) {
    printf("Error executing threads %s\n", cudaGetErrorString(err));
    exit(1);
  }
  if (cudaThreadSynchronize() != cudaSuccess) {
    printf("Error return from synchronizing\n");
    exit(1);
  }
  
  sizetmp = sizeof(float) * width;
  if (cudaMemcpy2D(bpgpu, sizetmp, devSlice, slicePitch, sizetmp, thick, 
                   //if (cudaMemcpy(bpgpu, devSlice, sizetmp * thick, 
                   cudaMemcpyDeviceToHost) != cudaSuccess) {
    printf("Error copying slice back to host\n");
    exit(1);
  }

#ifndef REPRODUCE_TEST
  printf("GPU 1D projection walltime %.4f sec\n", walltime() - tstart);
#endif
  minMaxDiff(bpcpu, bpgpu, projarr, width * thick, kstart, support, width);

  
  // Loop on 2D kernel calls
  cudaMemset2D(devSlice, slicePitch, 0, width * sizeof(float), thick);
  tstart = walltime();
  dim3 blockSize2(16, 16, 1);
  dim3 gridSize2((width + blockSize2.x - 1) / blockSize2.x, 
                 (thick + blockSize2.y - 1) / blockSize2.y, 1);
  for (k = kstart; k <= kend; k++) {
    theta = DTOR * (k - numproj / 2) * anginc;
    costh = cos(theta);
    sinth = sin(theta);
    xx = k;
    xstart = - costh * width / 2. + support / 2.;
    bpLine2D<<<gridSize2, blockSize2>>>(devSlice, slicePitch / 4, width, thick, xx,
                                        xstart, costh, sinth);
  }
  err = cudaGetLastError();
  if (err != cudaSuccess) {
    printf("Error executing threads %s\n", cudaGetErrorString(err));
    exit(1);
  }
  if (cudaThreadSynchronize() != cudaSuccess) {
    printf("Error return from synchronizing\n");
    exit(1);
  }
  
  sizetmp = sizeof(float) * width;
  if (cudaMemcpy2D(bpgpu, sizetmp, devSlice, slicePitch, sizetmp, thick, 
                   //if (cudaMemcpy(bpgpu, devSlice, sizetmp * thick, 
                   cudaMemcpyDeviceToHost) != cudaSuccess) {
    printf("Error copying slice back to host\n");
    exit(1);
  }

#ifndef REPRODUCE_TEST
  printf("GPU 2D projection walltime %.4f sec\n", walltime() - tstart);
#endif
  minMaxDiff(bpcpu, bpgpu, projarr, width * thick, kstart, support, width);

  cudaMemcpyToSymbol(cosines, cosarr, numproj * sizeof(float));
  cudaMemcpyToSymbol(sines, sinarr, numproj * sizeof(float));

  // Do all angles at once
  cudaMemset2D(devSlice, slicePitch, 0, width * sizeof(float), thick);
  tstart = walltime();
  bpAllAngles<<<gridSize2, blockSize2>>>(devSlice, slicePitch / 4, width, thick, 
                                         numproj, support);
  err = cudaGetLastError();
  if (err != cudaSuccess) {
    printf("Error executing threads %s\n", cudaGetErrorString(err));
    exit(1);
  }
  if (cudaThreadSynchronize() != cudaSuccess) {
    printf("Error return from synchronizing\n");
    exit(1);
  }
  
  sizetmp = sizeof(float) * width;
  if (cudaMemcpy2D(bpgpu, sizetmp, devSlice, slicePitch, sizetmp, thick, 
                   //if (cudaMemcpy(bpgpu, devSlice, sizetmp * thick, 
                   cudaMemcpyDeviceToHost) != cudaSuccess) {
    printf("Error copying slice back to host\n");
    exit(1);
  }

#ifndef REPRODUCE_TEST
  printf("GPU all angles walltime %.4f sec\n", walltime() - tstart);
#endif
  minMaxDiff(bpcpu, bpgpu, projarr, width * thick, kstart, support, width);


  cudaFree(devSlice);
  cudaFreeArray(devProj);
  free(bpcpu);
  free(bpgpu);
  free(projarr);
  return 0;
}

static void minMaxDiff(float *bpcpu, float *bpgpu, float *projarr, int num, int kstart,
                       int support, int width)
{
  int i, nsum = 0;
  float  diff, diffmax, cmin, cmax, gmin, gmax, fracmax, frac;
  double fracsum = 0.;
  cmin = 1.e38;
  cmax = -cmin;
  gmin = 1.e38;
  gmax = -cmin;
  diffmax = 0.;
  fracmax = 0.;
  for (i = 0; i < num; i++) {
    if (cmin > bpcpu[i])
      cmin = bpcpu[i];
    if (cmax < bpcpu[i])
      cmax = bpcpu[i];
    if (gmin > bpgpu[i])
      gmin = bpgpu[i];
    if (gmax < bpgpu[i])
      gmax = bpgpu[i];
    diff = fabs((double)(bpcpu[i] - bpgpu[i]));
    if (diffmax < diff)
      diffmax = diff;
    if (diff > 0) {
      frac = diff / (bpcpu[i] > bpgpu[i] ? bpcpu[i] : bpgpu[i]);
      if (fracmax < frac)
        fracmax = frac;
      fracsum += frac;
      nsum++;
    }
    /*if (i < width) printf("%.1f %.1f %.1f %.1f %.1f\n", 
                          projarr[(kstart - 1) * support + i + (support - width) / 2],
                          projarr[kstart * support + i + (support - width) / 2],
                          projarr[(kstart + 1) * support + i + (support - width) / 2],
                          bpcpu[i], bpgpu[i]);*/
  }
#ifdef REPRODUCE_TEST
  printf("Max difference %.3f\n", diffmax);
#else
  printf("CPU Min %.2f max %.2f GPU min %.2f max %.2f max diff %.2f\n"
         "frac diff mean %f  max %f\n", 
         cmin, cmax, gmin, gmax, diffmax, fracsum / nsum, fracmax);
#endif
}
